from typing import *
from collections.abc import Iterable

from ray.rllib.agents.callbacks import DefaultCallbacks
from ray.rllib.env import BaseEnv
from ray.rllib.evaluation import MultiAgentEpisode, RolloutWorker
from ray.rllib.policy import Policy


class TrainingCallbacks(DefaultCallbacks):
    """ 
    TrainingCallbacksCallbacks class. Contains 
    the definition of callbacks used during model training 
    """

    def on_episode_end(self, 
                       *, 
                       worker: RolloutWorker, 
                       base_env: BaseEnv,
                       policies: Dict[str, Policy], 
                       episode: MultiAgentEpisode,
                       env_index: int, 
                       **kwargs) -> None:
        """
        Runs when an episode is done.

        Args: -> check DefaultCallbacks class
        """

        #Info returned by the episode
        info = episode.last_info_for()
        if info is not None:
            if "custom_metrics" in info:
                for key, item in info["custom_metrics"].items():
                    episode.custom_metrics[key] = item


class epsConstraintCallbacks(TrainingCallbacks):
    """ 
    epsConstraintCallbacks class. Contains 
    the on_train_results callbacks used to decrease the
    constraint satisfaction tolerance during training
    """

    def on_train_result(self, 
                         *, 
                         trainer, 
                         result: dict, 
                         **kwargs) -> None:
        """
        Called at the end of Trainable.train().

        Args: -> check DefaultCallbacks class
        """

        training_iter = result["training_iteration"]
        
        trainer.workers.foreach_worker(
            lambda ev: ev.foreach_env(
                lambda env: env.set_cstr_tolerance(training_iter)))


class EvaluationCallbacks(DefaultCallbacks):
    """ 
    EvaluationCallbacksCallbacks class. Contains 
    the definition of callbacks used during model evaluation 
    """


    def on_episode_step(self, 
                        *, 
                        worker: RolloutWorker, 
                        base_env: BaseEnv,
                        episode: MultiAgentEpisode, 
                        env_index: int, 
                        **kwargs) -> None:
        """
        Runs on each episode step.

        Args: -> check DefaultCallbacks class
        """

        #Info and done returned by the episode
        info = episode.last_info_for()

        if info is not None:
            if "episode_step_data" in info:
                for key, item in info["episode_step_data"].items():
                    if isinstance(item, Iterable):
                        if episode.length == 1:
                            episode.user_data[key] = list(item)
                        else:
                            episode.user_data[key] = episode.user_data[key] + list(item)
                    else:
                        if episode.length == 1:
                            episode.user_data[key] = [item]
                        else:
                            episode.user_data[key].append(item)
    

    def on_episode_end(self, 
                       *, 
                       worker: RolloutWorker, 
                       base_env: BaseEnv,
                       policies: Dict[str, Policy], 
                       episode: MultiAgentEpisode,
                       env_index: int, 
                       **kwargs) -> None:
        """
        Runs when an episode is done.

        Args: -> check DefaultCallbacks class
        """

        #Info and done returned by the episode
        info = episode.last_info_for()

        if info is not None:
            if "episode_step_data" in info:
                for key in info["episode_step_data"].keys():
                    episode.hist_data[key] = episode.user_data[key]
                    episode.hist_data[key + "_length"] = [len(episode.user_data[key])]
            if "episode_end_data" in info:
                for key, item in info["episode_end_data"].items():
                    if isinstance(item, Iterable):
                        episode.hist_data[key] = [item[-1]]
                    else:
                        episode.hist_data[key] = [item]
            if "custom_metrics" in info:
                for key, item in info["custom_metrics"].items():
                    if isinstance(item, Iterable):
                        episode.hist_data[key] = [item[-1]]
                        episode.custom_metrics[key] = item[-1]
                    else:
                        episode.hist_data[key] = [item]
                        episode.custom_metrics[key] = item





from pyrlprob.problem import RLProblem
from pyrlprob.tests import *

from typing import *
import matplotlib.pyplot as plt
import yaml
import os
import matplotlib

from pyrlprob.utils.plots import plot_metric


def test_train_py(model: str="fcnet",
        res_dir: Optional[str]=None) -> None:
    """
    Test pyrlprob training functionalities in the Landing1D environment
        written in python.

    Args:
        res_dir: path where results are saved. Current directory if not specified.
    """

    #Train
    _, _, _, _ = \
        test_landing_env_train(py_or_cpp="py",
                               model=model,
                               res_dir=res_dir)


def test_train_cpp(res_dir: Optional[str]=None) -> None:
    """
    Test pyrlprob training functionalities in the Landing1D environment
        written in c++.

    Args:
        res_dir: path where results are saved. Current directory if not specified.
    """

    #Train
    _, _, _, _ = \
        test_landing_env_train(py_or_cpp="cpp",
                               res_dir=res_dir)


def test_train_eval_py(model: str="fcnet",
                       res_dir: Optional[str]=None) -> None:
    """
    Test pyrlprob training, evaluation and post-processing functionalities 
        in the Landing1D environment written in python.

    Args:
        res_dir: path where results are saved. Current directory if not specified.
    """

    #Train, evaluate, post-process
    test_landing_env_train_eval(py_or_cpp="py",
                               model=model,
                               res_dir=res_dir)


def test_train_eval_cpp(res_dir: Optional[str]=None) -> None:
    """
    Test pyrlprob training, evaluation and post-processing functionalities 
        in the Landing1D environment written in c++.

    Args:
        res_dir: path where results are saved. Current directory if not specified.
    """

    #Train, evaluate, post-process
    test_landing_env_train_eval(py_or_cpp="cpp",
                               res_dir=res_dir)


def make_cpp_lib() -> None:
    """
    Create the dynamic library with the environments in c++.
    """

    __location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
    os.system("make -f " + os.path.join(__location__, "Makefile"))

    return


def test_landing_env_train(py_or_cpp: str="py",
                           model: str="fcnet",
                           res_dir: Optional[str]=None) -> Tuple[RLProblem, str, List[str], List[int]]:
    """
    Test pyrlprob training functionalities in the Landing1D environment.

    Args:
        py_or_cpp: python or cpp version of the environment?
        res_dir: path where results are saved. Current directory if not specified.
    """

    #Config file
    __location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))
    if py_or_cpp == "py":
        if model == "mlp":
            config = os.path.join(__location__, "landing1d_py.yaml")
        elif model == "lstm":
            config = os.path.join(__location__, "landing1d_lstm_py.yaml")
        elif model == "gtrxl":
            config = os.path.join(__location__, "landing1d_gtrxl_py.yaml")
        else:
            assert False, "Model not implemented."
    elif py_or_cpp == "cpp":
        if not any(fname.endswith('.so') for fname in os.listdir(__location__ + "/cpp_tests/")):
            assert False, "Run 'make_cpp_lib()' first to create the dynamic library, then launch again python."
        config = os.path.join(__location__, "landing1d_cpp.yaml")

    #Problem definition
    LandingProblem = RLProblem(config)

    #Training
    _, trainer_dir, exp_dirs, last_cps, _ = \
        LandingProblem.solve(res_dir, 
                             evaluate=False, 
                             postprocess=False,
                             debug=True)

    return LandingProblem, trainer_dir, exp_dirs, last_cps


def test_landing_env_train_eval(py_or_cpp: str="py",
                                model: str="fcnet",
                                res_dir: Optional[str]=None) -> None:
    """
    Test pyrlprob training, evaluation and post-processing functionalities 
        in the Landing1D environment.

    Args:
        py_or_cpp: python or cpp version of the environment?
        res_dir: path where results are saved. Current directory if not specified.
    """


    __location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))

    #Train
    LandingProblem, trainer_dir, exp_dirs, last_cps = \
        test_landing_env_train(py_or_cpp=py_or_cpp,
                               model=model,
                               res_dir=res_dir)

    #Create new config file for model re-training
    load = {"load": {"trainer_dir": trainer_dir,
            "prev_exp_dirs": exp_dirs,
            "prev_last_cps": last_cps}}
    stop = {"stop": {"training_iteration": 2*last_cps[-1]}}
    config_new_dict = {**LandingProblem.input_config, **stop, **load}
    with open(os.path.join(__location__, "landing1d_load.yaml"), 'w') as outfile:
        yaml.dump(config_new_dict, outfile)
    
    #Training, evaluation and postprocessing of pre-trained model
    config_new = os.path.join(__location__, "landing1d_load.yaml")
    LandingProblemPretrained = RLProblem(config_new)
    _, trainer_dir, exp_dirs, last_cps, _ = \
        LandingProblemPretrained.solve(debug=False)

    # #Plot of metric trend
    # plt.style.use("seaborn")
    # plt.style.use("seaborn")
    # matplotlib.rc('font', size=20)
    # fig = plot_metric("episode_reward",
    #                   exp_dirs,
    #                   last_cps)
    # plt.xlabel('training iteration', fontsize=20)
    # plt.ylabel('episode reward', fontsize=20)
    # fig.savefig(trainer_dir + "episode_reward.png")


if __name__ == "__main__":
    import pytest
    import sys
    sys.exit(pytest.main(["-v", __file__]))


import numpy as np
from typing import *
import importlib
import yaml
import copy

from ray.rllib.models import ModelCatalog

from pyrlprob.base_funs import *
from pyrlprob.utils import update
from pyrlprob.models import models


class RLProblem:
    """
    Abstract class defining a generic Reinforcement Learning Problem
    """

    def __init__(self, 
                 config_file: Union[str, Dict[str, Any]]) -> None:
        """ 
        Class constructor 
        
        Args:
            config_file (str or dict): name of the config file (.yaml) that constains the trainer, environment and
                post-processing settings or the dictionary with all the information
        """

        #Open config file
        if isinstance(config_file, str):
            settings = yaml.safe_load(open(config_file))
        else:
            settings = config_file
        self.input_config = settings

        #Trainer definition
        if isinstance(settings["run"], dict):
            self.__method = settings["run"]["method"]
            self.__algorithm = settings["run"]["algorithm"].upper()
            default_config = self.__algorithm + "_DEFAULT_CONFIG"
        else:
            self.__method = settings["run"]
            self.__algorithm = self.__method.upper()
            default_config = "DEFAULT_CONFIG"
        alg_module = importlib.import_module("ray.rllib.agents." + self.__method)
        self.trainer = getattr(alg_module, self.__algorithm + "Trainer")

        #Stopping criteria definition
        self.stop = settings["stop"]

        #Config definition
        self.config = getattr(alg_module, default_config).copy()
        update(self.config, settings["config"])

        #Evironment definition
        if "." in self.config["env"]:
            mod_name, env_name = self.config["env"].rsplit('.',1)
            mod = importlib.import_module(mod_name)
            env = getattr(mod, env_name)
            tune.register_env(self.config["env"], lambda config: env(config))
        self.env = self.config["env"]
        self.env_config = self.config["env_config"]

        #Model definition
        self.model = self.config["model"]
        if self.model["custom_model"] is not None:
            if self.model["custom_model"] in ["FCModelforRNNs", "MLPplusLSTM", \
                                              "MLPplusGTrXL"]:
                custom_model = getattr(models, self.model["custom_model"])
                self.model["custom_model"] = "pyrlprob.models." + self.model["custom_model"]

            if "." in self.model["custom_model"]:
                mod_name, model_name = self.model["custom_model"].rsplit('.',1)
                mod = importlib.import_module(mod_name)
                custom_model = getattr(mod, model_name)
                ModelCatalog.register_custom_model(self.model["custom_model"], custom_model)
        
        #Framework definition
        self.framework = self.config["framework"]

        #Gamma definition
        self.gamma = self.config["gamma"]

        #Evaluation config definition
        self.evaluation = bool(self.config["evaluation_interval"])
        self.evaluation_config = copy.deepcopy(self.config["evaluation_config"])
        if self.config["custom_eval_function"] is not None:
            mod_name, fun_name = self.config["custom_eval_function"].rsplit('.',1)
            mod = importlib.import_module(mod_name)
            self.config["custom_eval_function"] = getattr(mod, fun_name)
        self.custom_eval_function = self.config["custom_eval_function"]

        #Final evaluation
        self.final_eval_episodes = 1
        self.final_evaluation_config = {}
        if "final_evaluation" in settings:
            if settings["final_evaluation"]:
                if "final_eval_episodes" in settings:
                    self.final_eval_episodes = settings["final_eval_episodes"]
                if "final_evaluation_config" in settings:
                    self.final_evaluation_config = settings["final_evaluation_config"]
        if self.evaluation_config is not None and self.evaluation_config != {}:
            update(self.evaluation_config, self.final_evaluation_config)
        else:
            self.evaluation_config = self.final_evaluation_config
        if "record_env" in self.evaluation_config:
            if self.evaluation_config["record_env"] and \
                isinstance(self.evaluation_config["record_env"], str):
                os.makedirs(self.evaluation_config["record_env"], exist_ok=True)

        #Custom metrics
        self.custom_metrics = []
        if "custom_metrics" in settings:
            self.custom_metrics = settings["custom_metrics"]

        #Callbacks definition
        if not callable(self.config["callbacks"]):
            if self.config["callbacks"] != "DefaultCallbacks":
                if self.config["callbacks"] in ["TrainingCallbacks", "epsConstraintCallbacks"]:
                    self.config["callbacks"] = getattr(callbacks, self.config["callbacks"])
                else:
                    mod_name, fun_name = self.config["callbacks"].rsplit('.',1)
                    mod = importlib.import_module(mod_name)
                    self.config["callbacks"] = getattr(mod, fun_name)
        
        #Postprocessing definition
        self.postproc_data = {"custom_metrics": self.custom_metrics}
        if "postproc_data" in settings:
            self.postproc_data["episode_step_data"] = []
            self.postproc_data["episode_end_data"] = []
            if "episode_step_data" in settings["postproc_data"]:
                self.postproc_data["episode_step_data"] = settings["postproc_data"]["episode_step_data"]
            if "episode_end_data" in settings["postproc_data"]:
                self.postproc_data["episode_end_data"] = settings["postproc_data"]["episode_end_data"]
        
        #Pre-trained model definition
        self.load = None
        if "load" in settings:
            self.load = {}
            self.load["logdir"] = settings["load"]["trainer_dir"]
            self.load["exp_dirs"] = settings["load"]["prev_exp_dirs"]
            self.load["last_cps"] = settings["load"]["prev_last_cps"]
            _, self.load["checkpoint_dir"] = \
                get_cp_dir_and_model(self.load["exp_dirs"][-1], self.load["last_cps"][-1])
    

    def solve(self,
              logdir: Optional[str]=None,
              num_cp_to_keep: str="all",
              evaluate: bool=True, 
              best_metric: str="episode_reward_mean",
              min_or_max: str="max",
              postprocess: bool=True,
              debug: bool=False,
              open_ray: bool=True,
              return_time: bool=False) ->  \
                  Union[Tuple[Dict[str, Any], str, List[str], List[int], str, float], \
                        Tuple[Dict[str, Any], str, List[str], List[int], str]]:
        """
        Solve a RL problem.
        It include pre-processing and training, 
        and may include evaluation and post-processing.

        Args:
            logdir (str): name of the directory where training results are saved
            num_cp_to_keep (str): number of checkpoints to keep. If "all", all checkpoints are kept, if "best", only the best checkpoint is kept
            evaluate (bool): whether to do evaluation
            best_metric (str): metric to be used to determine the best checkpoint in exp_dir during evaluation.
            min_or_max (str): if best_metric must be minimized or maximized
            postprocess (bool): whether to do postprocessing
            debug (bool): whether to print worker's logs.
            open_ray (bool): whether to open/close ray
            return_time (bool): whether to return run time per iter
        
        Return:
            best_result (str): best result of the experiment
            trainer_dir (str): trainer directory
            exp_dirs (list): experiment directories
            last_cps (list): last checkpoints of the experiments
            best_cp_dir (str): best checkpoint directory
            (optional) run_time (float): total run time
        """
        
        #Training
        if return_time:
            best_result, trainer_dir, best_exp_dir, last_checkpoint, run_time = training(trainer=self.trainer, 
                                                                            config=self.config, 
                                                                            stop=self.stop,
                                                                            best_metric=best_metric,
                                                                            min_or_max=min_or_max,
                                                                            num_cp_to_keep=num_cp_to_keep,
                                                                            evaluation_active=self.evaluation,
                                                                            logdir=logdir,
                                                                            load=self.load,
                                                                            debug=debug,
                                                                            open_ray=open_ray,
                                                                            return_time=return_time)
        else:
            best_result, trainer_dir, best_exp_dir, last_checkpoint = training(trainer=self.trainer, 
                                                                  config=self.config, 
                                                                  stop=self.stop,
                                                                  best_metric=best_metric,
                                                                  min_or_max=min_or_max,
                                                                  num_cp_to_keep=num_cp_to_keep,
                                                                  evaluation_active=self.evaluation,
                                                                  logdir=logdir,
                                                                  load=self.load,
                                                                  debug=debug,
                                                                  open_ray=open_ray,
                                                                  return_time=return_time)
        
        #Save config file
        with open(best_exp_dir + "config.yaml", 'w') as outfile:
            yaml.dump(self.input_config, outfile)
        
        #Evaluation and Postprocessing
        if evaluate:
            if num_cp_to_keep == "best":
                best_metric = last_checkpoint
            exp_dirs, last_cps, best_cp_dir = self.evaluate(trainer_dir=trainer_dir,
                                                            exp_dir=best_exp_dir,
                                                            last_checkpoint=last_checkpoint,
                                                            best_metric=best_metric,
                                                            min_or_max=min_or_max,
                                                            do_postprocess=postprocess,
                                                            debug=debug)
        else:
            exp_dirs = [best_exp_dir]
            last_cps = [last_checkpoint]
            if self.load is not None:
                exp_dirs = self.load["exp_dirs"] + exp_dirs
                last_cps = self.load["last_cps"] + last_cps
            best_cp_dir, _ = get_cp_dir_and_model(best_exp_dir, last_checkpoint)

        if return_time:
            return best_result, trainer_dir, exp_dirs, last_cps, best_cp_dir, run_time
        else:
            return best_result, trainer_dir, exp_dirs, last_cps, best_cp_dir


    def evaluate(self,
                 trainer_dir: Optional[str] = None,
                 exp_dir: Optional[str] = None,
                 last_checkpoint: Optional[int] = None,
                 best_metric: str = "episode_reward_mean",
                 min_or_max: str = "max",
                 do_postprocess: Optional[bool] = True,
                 debug: bool = False) -> Tuple[List[str], List[int], str]:
        """
        Evaluate current model.
        It may include postprocessing.

        Args:
            trainer_dir (str): trainer directory
            exp_dir (str): experiment directory
            last_checkpoint (int): last checkpoint of the experiment
            best_metric (str): metric to be used to determine the best checkpoint in exp_dir
            min_or_max (str): if best_metric must be minimized or maximized
            do_postprocess (bool): whether to do postprocessing
            debug (bool): is debugging mode on?
        """

        exp_dirs = [exp_dir]
        last_cps = [last_checkpoint]
        if self.load is not None:
            exp_dirs = self.load["exp_dirs"] + exp_dirs
            last_cps = self.load["last_cps"] + last_cps
        best_cp_dir = evaluation(trainer=self.trainer,
                                 trainer_dir=trainer_dir, 
                                 exp_dirs=exp_dirs,
                                 last_cps=last_cps,
                                 config=self.config,
                                 env_name=self.env, 
                                 env_config=self.env_config,
                                 evaluation_num_episodes=self.final_eval_episodes,
                                 evaluation_config=self.evaluation_config, 
                                 custom_eval_function=self.custom_eval_function, 
                                 best_metric=best_metric,
                                 min_or_max=min_or_max,
                                 metrics_and_data=self.postproc_data, 
                                 is_evaluation_env=self.evaluation, 
                                 do_postprocess=do_postprocess,
                                 debug=debug)
        
        return exp_dirs, last_cps, best_cp_dir
    

    def postprocess(self,
                    exp_dir: str,
                    checkpoint: int) -> str:
        """
        Postprocess the experiment.

        Args:
            exp_dir (str): experiment directory
            checkpoint (int): experiment's checkpoint to postprocess
        
        Returns:
            cp_dir (str): postprocessed checkpoint directory
        """

        cp_dir = postprocessing(best_exp_dir=exp_dir, 
                                checkpoint=checkpoint, 
                                metrics_and_data=self.postproc_data, 
                                is_evaluation_env=self.evaluation)
        
        return cp_dir



from typing import *
import numpy as np
import matplotlib.pyplot as plt

from pyrlprob.utils.auxiliary import moving_average, metric_training_trend

def plot_metric(metric_name: str,
                experiment_dirs: List[str],
                last_checkpoints: List[int],
                window: int = 1,
                step: int = 1,
                factor: float = 1.,
                fig: Optional[Callable] = None,
                label: Optional[str] = None,
                color: Optional[str] = None) -> Callable:
    """
    plot the trend during training of a metric

    Args:
        metric_name (str): full name of the metric (with the path)
        experiment_dirs (list): list of the experiment directories
        last_checkpoints (list): list with the last checkpoints of the experiments
        window (int): window to evaluate the moving average
        step (int): plot points every step steps
        factor (float): factor to multiply the metric values
        fig (callable): figure object
        label (str): the label of the plotted curve
        color (str): color of the curve (if None -> random color)
    
    Return:
        fig: figure object
    """

    #Retrieve metric min, mean and max trend
    metric_min = metric_training_trend(metric_name + "_min",
                                    experiment_dirs,
                                    last_checkpoints)
    metric_mean = metric_training_trend(metric_name + "_mean",
                                        experiment_dirs,
                                        last_checkpoints)
    metric_max = metric_training_trend(metric_name + "_max",
                                    experiment_dirs,
                                    last_checkpoints)

    #Training iterations
    training_iter = [i for i in range(len(metric_mean))]
    
    #Evaluate the moving average
    metric_mean_mov = moving_average(metric_mean, window=window)
    training_iter_mov = training_iter[len(training_iter) - len(metric_mean_mov):]
    if metric_min != None:
        metric_min_mov = moving_average(metric_min, window=window)
        metric_max_mov = moving_average(metric_max, window=window)
        metric_std_min_mov = np.array(metric_mean_mov) - abs(np.array(metric_mean_mov) - np.array(metric_min_mov))/4.
        metric_std_max_mov = np.array(metric_mean_mov) + abs(np.array(metric_mean_mov) - np.array(metric_max_mov))/4.

    #Plot
    if fig is None:
        fig = plt.figure()
        fig.set_size_inches(9.7,6.4)
    ax = fig.gca()
    ax.plot(training_iter_mov[::step], metric_mean_mov[::step]*factor, '-', linewidth='2.5', color=color, label=label)
    if metric_min !=  None:
        plt.fill_between(training_iter_mov[::step], metric_std_min_mov[::step]*factor, metric_std_max_mov[::step]*factor, alpha=0.3, color=color)
    
    return fig


import logging

from ray.rllib.models.tf.misc import normc_initializer
from ray.rllib.models.tf.tf_modelv2 import TFModelV2
from ray.rllib.models.modelv2 import ModelV2
from ray.rllib.utils.framework import try_import_tf
from typing import Dict, List
from ray.rllib.utils.typing import TensorType, List, ModelConfigDict
from ray.rllib.models.utils import get_activation_fn
from ray.rllib.utils.annotations import override
from ray.rllib.utils.tf_ops import one_hot
from ray.rllib.models.tf.recurrent_net import RecurrentNetwork
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.policy.view_requirement import ViewRequirement
from ray.rllib.models.tf.layers import (
    GRUGate,
    RelativeMultiHeadAttention,
    SkipConnection,
)
from ray.rllib.models.tf.attention_net import PositionwiseFeedforward
import gym
from gym.spaces import Box, Discrete, MultiDiscrete
import numpy as np

tf1, tf, tfv = try_import_tf()
logger = logging.getLogger(__name__)


class FCModelforRNNs(TFModelV2):
    """Custom FC model to be used before an RNN/attention wrapper for policy gradient algorithms.
    The model is the same as the default FCModel, except that the last input observation is
    interpreted as the previous "estimated" reward and is directly concatenated
    to the output of the last hidden layer. This is done to allow the RNN/attention
    wrapper to learn a policy that is conditioned on the previous reward.
    """

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
    ):
        super(FCModelforRNNs, self).__init__(
            obs_space, action_space, num_outputs, model_config, name
        )

        hiddens = list(model_config.get("fcnet_hiddens", []))
        activation = model_config.get("fcnet_activation")
        activation = get_activation_fn(activation)
        
        # We are using obs_flat, so take the flattened shape as input, minus one (the reward)
        inputs = tf.keras.layers.Input(
            shape=(int(np.product(obs_space.shape) - 1),), name="observations"
        )
        # Last hidden layer output (before logits outputs).
        last_layer = inputs
        # The action distribution outputs.
        logits_out = None
        i = 1

        # Create layers 0 to second-last.
        for size in hiddens[:-1]:
            last_layer = tf.keras.layers.Dense(
                size,
                name="fc_{}".format(i),
                activation=activation,
                kernel_initializer=normc_initializer(1.0),
            )(last_layer)
            i += 1

        # Finish the layers with the provided sizes (`hiddens`), plus -
        # iff num_outputs > 0 - a last linear layer of size num_outputs.
        if len(hiddens) > 0:
            last_layer = tf.keras.layers.Dense(
                hiddens[-1],
                name="fc_{}".format(i),
                activation=activation,
                kernel_initializer=normc_initializer(1.0),
            )(last_layer)
        if num_outputs:
            logits_out = tf.keras.layers.Dense(
                num_outputs,
                name="fc_out",
                activation=None,
                kernel_initializer=normc_initializer(0.01),
            )(last_layer)
        # Adjust num_outputs to be the number of nodes in the last layer.
        else:
            self.num_outputs = ([int(np.product(obs_space.shape))] + hiddens[-1:])[
                -1
            ]

        last_vf_layer = None

        value_out = tf.keras.layers.Dense(
            1,
            name="value_out",
            activation=None,
            kernel_initializer=normc_initializer(0.01),
        )(last_vf_layer if last_vf_layer is not None else last_layer)

        self.base_model = tf.keras.Model(
            inputs, [(logits_out if logits_out is not None else last_layer), value_out]
        )

    def forward(
        self,
        input_dict: Dict[str, TensorType],
        state: List[TensorType],
        seq_lens: TensorType,
    ) -> tuple[TensorType, List[TensorType]]:
        
        # Remove the last element from the observation
        input_obs_reduced = input_dict["obs_flat"][:, :-1]

        # Output of the fc model
        model_out, self._value_out = self.base_model(input_obs_reduced)

        # Last element of the observation (the reward)
        last_obs = tf.reshape(
            tf.cast(input_dict["obs_flat"][:, -1], tf.float32), [-1, 1]
        )

        # Previous reward
        prev_r = [last_obs]
        
        # New output
        model_out = tf.concat([model_out] + prev_r, axis=1)

        return model_out, state

    def value_function(self) -> TensorType:
        return tf.reshape(self._value_out, [-1])


class MLPplusLSTM(RecurrentNetwork):
    """LSTM-based model for meta-RL. The policy model is made up of an MLP encoder and an LSTM cell.
     The encoder is a simple feedforward network that preprocesses the observation before sending
     it to the LSTM cell, possible appending also the previous action and reward.
     The LSTM cell is a standard LSTM cell that outputs the logits.
     The value function estimate is computed either by a separate feedforward network that takes the
     same observation as input, or is an additional output of the LSTM cell.

     Args:
        custom_model_kwargs:
            lstm_cell_size (int): The size of the LSTM cell.
            encoder_hiddens (list): The sizes of the hidden layers of the encoder.
            encoder_activation (str): The activation function of the encoder.
            vf_hiddens (list): The sizes of the hidden layers of the value function.
            vf_activation (str): The activation function of the value function.
            lstm_use_prev_action (bool): Whether to use the previous action as input to the LSTM cell.
            lstm_use_prev_reward (bool): Whether to use the previous reward as input to the LSTM cell.
        model_config:
            max_seq_len (int): The maximum length of the input observation sequence.
            vf_share_layers (bool): Whether to share the layers of the value function with the policy.
            free_log_std (bool): Whether to generate free-floating bias variables for the second half of the model outputs.
      """

    def __init__(
        self,
        obs_space,
        action_space,
        num_outputs,
        model_config,
        name,
        **custom_model_kwargs
    ):
        super(MLPplusLSTM, self).__init__(
            obs_space, action_space, num_outputs, model_config, name
        )
        
        if logger.isEnabledFor(logging.INFO):
            print("custom_model_kwargs: {}".format(custom_model_kwargs))

        # Encoder parameters
        self.encoder_hiddens = list(custom_model_kwargs.get("encoder_hiddens", []))
        self.encoder_activation = custom_model_kwargs.get("encoder_activation")
        encoder_activation = get_activation_fn(self.encoder_activation)

        # LSTM parameters
        self.cell_size = custom_model_kwargs.get("lstm_cell_size")
        self.use_prev_action = custom_model_kwargs.get("lstm_use_prev_action")
        self.use_prev_reward = custom_model_kwargs.get("lstm_use_prev_reward")
        self.max_seq_len = model_config.get("max_seq_len")
        free_log_std = model_config.get("free_log_std")

        # Value function parameters
        self.vf_share_layers = model_config.get("vf_share_layers")
        if not self.vf_share_layers:
            self.vf_hiddens = list(custom_model_kwargs.get("vf_hiddens", []))
            self.vf_activation = custom_model_kwargs.get("vf_activation")
            vf_activation = get_activation_fn(self.vf_activation)
        
        # Number of outputs (logit dim)
        self.num_outputs = num_outputs

        # Generate free-floating bias variables for the second half of
        # the outputs, if free_log_std = True
        if free_log_std:
            assert num_outputs % 2 == 0, (
                "num_outputs must be divisible by two",
                num_outputs,
            )
            num_outputs = num_outputs // 2
            self.log_std_var = tf.Variable(
                [0.0] * num_outputs, dtype=tf.float32, name="log_std"
            )
        
        # Action dimension
        if isinstance(action_space, Discrete):
            self.action_dim = action_space.n
        elif isinstance(action_space, MultiDiscrete):
            self.action_dim = np.sum(action_space.nvec)
        elif action_space.shape is not None:
            self.action_dim = int(np.product(action_space.shape))
        else:
            self.action_dim = int(len(action_space))

        ## DEFINE ENCODER LAYERS ##
            
        # Input layer
        input_layer_encoder = tf.keras.layers.Input(
            shape=(int(np.product(obs_space.shape)), ), name="observations"
        )

        # Encoder hidden layers
        last_layer_encoder = input_layer_encoder
        i = 1
        for size in self.encoder_hiddens:
            last_layer_encoder = tf.keras.layers.Dense(
                size,
                name="fc_encoder_{}".format(i),
                activation=encoder_activation,
                kernel_initializer=normc_initializer(1.0),
            )(last_layer_encoder)
            i += 1

        # Number of outputs (logit dim)
        self.encoder_out_size = self.encoder_hiddens[-1]

        ## DEFINE LSTM LAYERS ##

        # Add prev-action/reward nodes to input to LSTM.
        self.num_lstm_inputs = self.encoder_out_size
        if self.use_prev_action:
            self.num_lstm_inputs += self.action_dim
        if self.use_prev_reward:
            self.num_lstm_inputs += 1

        # Input layer
        input_layer_lstm = tf.keras.layers.Input(
            shape=(None, self.num_lstm_inputs), name="inputs_lstm"
        )
        state_in_h = tf.keras.layers.Input(shape=(self.cell_size,), name="h")
        state_in_c = tf.keras.layers.Input(shape=(self.cell_size,), name="c")
        seq_in = tf.keras.layers.Input(shape=(), name="seq_in", dtype=tf.int32)

        # LSTM cell
        lstm_out, state_h, state_c = tf.keras.layers.LSTM(
            self.cell_size, return_sequences=True, return_state=True, name="lstm"
        )(
            inputs=input_layer_lstm,
            mask=tf.sequence_mask(seq_in),
            initial_state=[state_in_h, state_in_c],
        )

        # Output layer
        logits = tf.keras.layers.Dense(
            num_outputs, activation=tf.keras.activations.linear, name="logits"
        )(lstm_out)

        # Concat the log std vars to the end of the state-dependent means.
        if free_log_std:

            def tiled_log_std(x):
                return tf.tile(tf.expand_dims(tf.expand_dims(self.log_std_var, 0),1), \
                               [tf.shape(x)[0], tf.shape(x)[1], 1])

            log_std_out = tf.keras.layers.Lambda(tiled_log_std)(input_layer_lstm)
            logits = tf.keras.layers.Concatenate(axis=2)([logits, log_std_out])

        ## DEFINE VALUE FUNCTION LAYERS ##
            
        # Value function hiddens layers
        if not self.vf_share_layers:
            last_vf_layer = input_layer_encoder
            i = 1
            for size in self.vf_hiddens:
                last_vf_layer = tf.keras.layers.Dense(
                    size,
                    name="fc_value_{}".format(i),
                    activation=vf_activation,
                    kernel_initializer=normc_initializer(1.0),
                )(last_vf_layer)
                i += 1
        else:
            last_vf_layer = lstm_out
        
        # Output layer
        value_out = tf.keras.layers.Dense(
            1,
            name="value_out",
            activation=None,
            kernel_initializer=normc_initializer(0.01),
        )(last_vf_layer)

        ## CREATE THE MODELS ##

        # Encoder model
        if self.vf_share_layers:
            self.encoder_model = tf.keras.Model(
                inputs=input_layer_encoder, outputs=last_layer_encoder
            )
        else:
            self.encoder_model = tf.keras.Model(
                inputs=input_layer_encoder, outputs=[last_layer_encoder, value_out]
            )
        
        # LSTM model
        if self.vf_share_layers:
            self.rnn_model = tf.keras.Model(
                inputs=[input_layer_lstm, seq_in, state_in_h, state_in_c],
                outputs=[logits, value_out, state_h, state_c],
            )
        else:
            self.rnn_model = tf.keras.Model(
                inputs=[input_layer_lstm, seq_in, state_in_h, state_in_c],
                outputs=[logits, state_h, state_c],
            )

        # Print out model summary in INFO logging mode.
        # if logger.isEnabledFor(logging.INFO):
        self.encoder_model.summary()
        self.rnn_model.summary()

        # Add prev-a/r to this model's view, if required.
        if self.use_prev_action:
            self.view_requirements[SampleBatch.PREV_ACTIONS] = \
                ViewRequirement(SampleBatch.ACTIONS, space=self.action_space,
                                shift=-1)
        if self.use_prev_reward:
            self.view_requirements[SampleBatch.PREV_REWARDS] = \
                ViewRequirement(SampleBatch.REWARDS, shift=-1)


    @override(RecurrentNetwork)
    def forward(self, input_dict: Dict[str, TensorType],
                state: List[TensorType],
                seq_lens: TensorType) -> tuple[TensorType, List[TensorType]]:
        
        assert seq_lens is not None

        # Push observations through encoder net's `forward()` first.
        if self.vf_share_layers:
            wrapped_out = self.encoder_model(input_dict["obs_flat"])
        else:
            wrapped_out, self._value_out = self.encoder_model(input_dict["obs_flat"])

        # Concat. prev-action/reward if required.
        prev_a_r = []
        if self.use_prev_action:
            prev_a = input_dict[SampleBatch.PREV_ACTIONS]
            if isinstance(self.action_space, (Discrete, MultiDiscrete)):
                prev_a = one_hot(prev_a, self.action_space)
            prev_a_r.append(
                tf.reshape(tf.cast(prev_a, tf.float32), [-1, self.action_dim]))
        if self.use_prev_reward:
            prev_a_r.append(
                tf.reshape(
                    tf.cast(input_dict[SampleBatch.PREV_REWARDS], tf.float32),
                    [-1, 1]))

        if prev_a_r:
            wrapped_out = tf.concat([wrapped_out] + prev_a_r, axis=1)

        # Then through our LSTM.
        input_dict["obs_flat"] = wrapped_out

        return super().forward(input_dict, state, seq_lens)
    

    @override(RecurrentNetwork)
    def forward_rnn(self, inputs, state, seq_lens):

        if self.vf_share_layers:
            model_out, self._value_out, h, c = self.rnn_model([inputs, seq_lens] + state)
        else:
            model_out, h, c = self.rnn_model([inputs, seq_lens] + state)
        
        return model_out, [h, c]


    @override(ModelV2)
    def get_initial_state(self):
        return [
            np.zeros(self.cell_size, np.float32),
            np.zeros(self.cell_size, np.float32),
        ]


    @override(ModelV2)
    def value_function(self):
        return tf.reshape(self._value_out, [-1])


class MLPplusGTrXL(RecurrentNetwork):
    """GTrXL-based model for meta-RL. The policy model is made up of an encoder and a GTrXL net.
     The encoder is a simple feedforward network that preprocesses the observation before sending
     it to the GTrXL, plus, possibly, the previous n actions and rewards. The GTrXL then outputs the logits.
     The value function estimate is either computed by a separate feedforward network that takes the
     same observation as input, or is an additional output of the GTrXL.
    
     Args:
        custom_model_kwargs:
            num_transformer_units (int): The number of transformer units.
            attention_dim (int): The dimension of the transformer units.
            num_heads (int): The number of heads in the attention.
            head_dim (int): The dimension of each head.
            memory_inference (int): The number of previous outputs to use as memory input to each layer.
            position_wise_mlp_dim (int): The dimension of the position-wise MLP.
            init_gru_gate_bias (float): The initial bias of the GRU gate.
            encoder_hiddens (list): The sizes of the hidden layers of the encoder.
            encoder_activation (str): The activation function of the encoder.
            vf_hiddens (list): The sizes of the hidden layers of the value function.
            vf_activation (str): The activation function of the value function.
        model_config:
            max_seq_len (int): The maximum length of the input observation sequence.
            vf_share_layers (bool): Whether to share the layers of the value function with the policy.
            free_log_std (bool): Whether to generate free-floating bias variables for the second half of the model outputs.
    """

    def __init__(
        self,
        obs_space,
        action_space,
        num_outputs,
        model_config,
        name,
        **custom_model_kwargs
    ):
        super().__init__(
            obs_space, action_space, num_outputs, model_config, name
        )

        if logger.isEnabledFor(logging.INFO):
            print("custom_model_kwargs: {}".format(custom_model_kwargs))

        # Encoder parameters
        self.encoder_hiddens = list(custom_model_kwargs.get("encoder_hiddens", []))
        self.encoder_activation = custom_model_kwargs.get("encoder_activation")
        encoder_activation = get_activation_fn(self.encoder_activation)

        # GTrXL parameters
        self.num_transformer_units = custom_model_kwargs.get("num_transformer_units")
        self.attention_dim = custom_model_kwargs.get("attention_dim")
        self.num_heads = custom_model_kwargs.get("num_heads")
        self.head_dim = custom_model_kwargs.get("head_dim")
        self.memory_inference = custom_model_kwargs.get("memory_inference")
        self.memory_training = self.memory_inference
        self.position_wise_mlp_dim = custom_model_kwargs.get("position_wise_mlp_dim")
        self.init_gru_gate_bias = custom_model_kwargs.get("init_gru_gate_bias")
        self.use_n_prev_actions = custom_model_kwargs.get("use_n_prev_actions")
        self.use_n_prev_rewards = custom_model_kwargs.get("use_n_prev_rewards")
        self.max_seq_len = model_config.get("max_seq_len")
        free_log_std = model_config.get("free_log_std")
        
        # Value function parameters
        self.vf_share_layers = model_config.get("vf_share_layers")
        if not self.vf_share_layers:
            self.vf_hiddens = list(custom_model_kwargs.get("vf_hiddens", []))
            self.vf_activation = custom_model_kwargs.get("vf_activation")
            vf_activation = get_activation_fn(self.vf_activation)

        # Number of outputs (logit dim)
        self.num_outputs = num_outputs

        # Generate free-floating bias variables for the second half of
        # the outputs.
        if free_log_std:
            assert num_outputs % 2 == 0, (
                "num_outputs must be divisible by two",
                num_outputs,
            )
            num_outputs = num_outputs // 2
            self.log_std_var = tf.Variable(
                [0.0] * num_outputs, dtype=tf.float32, name="log_std"
            )

        # Action dimension
        if isinstance(action_space, Discrete):
            self.action_dim = action_space.n
        elif isinstance(action_space, MultiDiscrete):
            self.action_dim = np.sum(action_space.nvec)
        elif action_space.shape is not None:
            self.action_dim = int(np.product(action_space.shape))
        else:
            self.action_dim = int(len(action_space))

        ## DEFINE ENCODER LAYERS ##
            
        # Input layer
        input_layer_encoder = tf.keras.layers.Input(
            shape=(int(np.product(obs_space.shape)), ), name="observations"
        )

        # Encoder hidden layers
        last_layer_encoder = input_layer_encoder
        i = 1
        for size in self.encoder_hiddens:
            last_layer_encoder = tf.keras.layers.Dense(
                size,
                name="fc_encoder_{}".format(i),
                activation=encoder_activation,
                kernel_initializer=normc_initializer(1.0),
            )(last_layer_encoder)
            i += 1

        # Number of outputs (logit dim)
        self.encoder_out_size = self.encoder_hiddens[-1]

        ## DEFINE GTRXL LAYERS ##      

        # Add prev-action/reward nodes to input to GTrXL.
        self.num_gtrxl_inputs = self.encoder_out_size
        if self.use_n_prev_actions:
            self.num_gtrxl_inputs += self.use_n_prev_actions * self.action_dim
        if self.use_n_prev_rewards:
            self.num_gtrxl_inputs += self.use_n_prev_rewards

        # Input layer
        input_layer_gtrxl = tf.keras.layers.Input(
            shape=(None, self.num_gtrxl_inputs), name="inputs_gtrxl")
        memory_ins = [
            tf.keras.layers.Input(
                shape=(None, self.attention_dim),
                dtype=tf.float32,
                name="memory_in_{}".format(i))
            for i in range(self.num_transformer_units)
        ]
        E_out = tf.keras.layers.Dense(self.attention_dim)(input_layer_gtrxl)
        memory_outs = [E_out]

        # GTrXL layers
        for i in range(self.num_transformer_units):
            # RelativeMultiHeadAttention part.
            MHA_out = SkipConnection(
                RelativeMultiHeadAttention(
                    out_dim=self.attention_dim,
                    num_heads=self.num_heads,
                    head_dim=self.head_dim,
                    input_layernorm=True,
                    output_activation=tf.nn.relu),
                fan_in_layer=GRUGate(self.init_gru_gate_bias),
                name="mha_{}".format(i + 1))(
                    E_out, memory=memory_ins[i])
            # Position-wise MLP part.
            E_out = SkipConnection(
                tf.keras.Sequential(
                    (tf.keras.layers.LayerNormalization(axis=-1),
                     PositionwiseFeedforward(
                         out_dim=self.attention_dim,
                         hidden_dim=self.position_wise_mlp_dim,
                         output_activation=tf.nn.relu))),
                fan_in_layer=GRUGate(self.init_gru_gate_bias),
                name="pos_wise_mlp_{}".format(i + 1))(MHA_out)
            memory_outs.append(E_out)

        # Output layer
        logits = tf.keras.layers.Dense(
            num_outputs, activation=tf.keras.activations.linear, name="logits"
        )(E_out)

        # Concat the log std vars to the end of the state-dependent means.
        if free_log_std:

            def tiled_log_std(x):
                return tf.tile(tf.expand_dims(tf.expand_dims(self.log_std_var, 0),1), \
                               [tf.shape(x)[0], tf.shape(x)[1], 1])

            log_std_out = tf.keras.layers.Lambda(tiled_log_std)(input_layer_gtrxl)
            logits = tf.keras.layers.Concatenate(axis=2)([logits, log_std_out])
        
        ## DEFINE VALUE FUNCTION LAYERS ##
        
        # Value function hiddens layers
        if not self.vf_share_layers:
            last_vf_layer = input_layer_encoder
            i = 1
            for size in self.vf_hiddens:
                last_vf_layer = tf.keras.layers.Dense(
                    size,
                    name="fc_value_{}".format(i),
                    activation=vf_activation,
                    kernel_initializer=normc_initializer(1.0),
                )(last_vf_layer)
                i += 1
        else:
            last_vf_layer = E_out
        
        # Output layer
        value_out = tf.keras.layers.Dense(
            1,
            name="value_out",
            activation=None,
            kernel_initializer=normc_initializer(0.01),
        )(last_vf_layer)

        ## CREATE THE MODELS ##

        # Encoder model
        if self.vf_share_layers:
            self.encoder_model = tf.keras.Model(
                inputs=input_layer_encoder, outputs=last_layer_encoder
            )
        else:
            self.encoder_model = tf.keras.Model(
                inputs=input_layer_encoder, outputs=[last_layer_encoder, value_out]
            )

        # GTrXL model
        if self.vf_share_layers:
            outs = [logits, value_out]
        else:
            outs = [logits]
        self.gtrxl_model = tf.keras.Model(
            inputs=[input_layer_gtrxl] + memory_ins, outputs=outs + memory_outs[:-1])

        # Print out model summary in INFO logging mode.
        # if logger.isEnabledFor(logging.INFO):
        self.encoder_model.summary()
        self.gtrxl_model.summary()

        # Add prev steps to this model's view, if required.
        for i in range(self.num_transformer_units):
            space = Box(-1.0, 1.0, shape=(self.attention_dim, ))
            self.view_requirements["state_in_{}".format(i)] = \
                ViewRequirement(
                    "state_out_{}".format(i),
                    shift="-{}:-1".format(self.memory_inference),
                    # Repeat the incoming state every max-seq-len times.
                    batch_repeat_value=self.max_seq_len,
                    space=space)
            self.view_requirements["state_out_{}".format(i)] = \
                ViewRequirement(
                    space=space,
                    used_for_training=False)
        self.view_requirements["obs"].space = self.obs_space
        
        # Add prev-a/r to this model's view, if required.
        if self.use_n_prev_actions:
            self.view_requirements[SampleBatch.PREV_ACTIONS] = \
                ViewRequirement(
                    SampleBatch.ACTIONS,
                    space=self.action_space,
                    shift="-{}:-1".format(self.use_n_prev_actions))
        if self.use_n_prev_rewards:
            self.view_requirements[SampleBatch.PREV_REWARDS] = \
                ViewRequirement(
                    SampleBatch.REWARDS,
                    shift="-{}:-1".format(self.use_n_prev_rewards))


    @override(ModelV2)
    def forward(self, input_dict, state: List[TensorType],
                seq_lens: TensorType) -> tuple[TensorType, List[TensorType]]:
        
        assert seq_lens is not None

        # Push observations through encoder net's `forward()` first.
        if self.vf_share_layers:
            wrapped_out = self.encoder_model(input_dict["obs_flat"])
        else:
            wrapped_out, self._value_out = self.encoder_model(input_dict["obs_flat"])

        # Concat. prev-action/reward if required.
        prev_a_r = []
        if self.use_n_prev_actions:
            if isinstance(self.action_space, Discrete):
                for i in range(self.use_n_prev_actions):
                    prev_a_r.append(
                        one_hot(input_dict[SampleBatch.PREV_ACTIONS][:, i],
                                self.action_space))
            elif isinstance(self.action_space, MultiDiscrete):
                for i in range(
                        self.use_n_prev_actions,
                        step=self.action_space.shape[0]):
                    prev_a_r.append(
                        one_hot(
                            tf.cast(
                                input_dict[SampleBatch.PREV_ACTIONS]
                                [:, i:i + self.action_space.shape[0]],
                                tf.float32), self.action_space))
            else:
                prev_a_r.append(
                    tf.reshape(
                        tf.cast(input_dict[SampleBatch.PREV_ACTIONS],
                                tf.float32),
                        [-1, self.use_n_prev_actions * self.action_dim]))
        if self.use_n_prev_rewards:
            prev_a_r.append(
                tf.reshape(
                    tf.cast(input_dict[SampleBatch.PREV_REWARDS], tf.float32),
                    [-1, self.use_n_prev_rewards]))
        
        if prev_a_r:
            wrapped_out = tf.concat([wrapped_out] + prev_a_r, axis=1)
        
        # Then through the GTrXL
        input_dict["obs_flat"] = input_dict["obs"] = wrapped_out

        return self.forward_gtrxl(input_dict, state, seq_lens)


    def forward_gtrxl(self, input_dict, state, seq_lens):
        
        # Add the time dim to observations.
        B = tf.shape(seq_lens)[0]
        observations = input_dict[SampleBatch.OBS]

        shape = tf.shape(observations)
        T = shape[0] // B
        observations = tf.reshape(observations,
                                  tf.concat([[-1, T], shape[1:]], axis=0))

        all_out = self.gtrxl_model([observations] + state)

        out = tf.reshape(all_out[0], [-1, self.num_outputs])
        if self.vf_share_layers:
            self._value_out = all_out[1]
            memory_outs = all_out[2:]
        else:
            memory_outs = all_out[1:]

        return out, [
            tf.reshape(m, [-1, self.attention_dim]) for m in memory_outs
        ]

    # @override(RecurrentNetwork)
    # def get_initial_state(self) -> List[np.ndarray]:
    #     return [
    #         tf.zeros(self.view_requirements["state_in_{}".format(i)].space.shape)
    #         for i in range(self.num_transformer_units)
    #     ]
    
    @override(RecurrentNetwork)
    def get_initial_state(self) -> List[np.ndarray]:
        return []

    @override(ModelV2)
    def value_function(self) -> TensorType:
        return tf.reshape(self._value_out, [-1])



import numpy as np
from typing import *


def rk4(f, y0, t_eval, data):

    n_eq = len(y0)
    t0 = t_eval[0]
    dt = t_eval[1] - t_eval[0]

    f0 = np.zeros(n_eq)
    f(t0, y0, f0, data)
    
    t1 = t0 + dt / 2.0
    y1 = y0 + dt * f0 / 2.0
    f1 = np.zeros(n_eq)
    f(t1, y1, f1, data)
    
    t2 = t0 + dt / 2.0
    y2 = y0 + dt * f1 / 2.0
    f2 = np.zeros(n_eq)
    f(t2, y2, f2, data)
    
    t3 = t0 + dt	
    y3 = y0 + dt * f2
    f3 = np.zeros(n_eq)
    f(t3, y3, f3, data)
    
    y = y0 + dt * (f0 + 2.0 * f1 + 2.0 * f2 + f3) / 6.0

    return y


def dynamics(t, s, s_dot, data):

    """
    System dynamics: vertical landing on planetary body
        with constant gravity g and thrust T
    """
    #State
    h = s[0]
    v = s[1]
    m = s[2]

    #Data
    g = data[0]
    T = data[1]
    c = data[2] 

    #Equations of motion
    h_dot = v
    v_dot = - g + T/m
    m_dot = - T/c

    s_dot[0] = h_dot
    s_dot[1] = v_dot
    s_dot[2] = m_dot



from pyrlprob.tests.py_tests import *
from pyrlprob.tests.test_landing1d import test_train_py, make_cpp_lib, test_train_cpp, test_train_eval_py, test_train_eval_cpp
from pyrlprob.tests.test_multiple_gym import test_multiple_gym_envs as test_gym_envs
try:
    from pyrlprob.tests.cpp_tests import *
except ImportError:
    pass

__all__ = [
    "test_train_py",
    "make_cpp_lib",
    "test_train_cpp",
    "test_train_eval_py",
    "test_train_eval_cpp",
    "test_gym_envs"
]


import numpy as np
import time
from typing import *
import importlib
import os
import multiprocessing

import ray
from ray import tune
from ray.rllib.env.vector_env import VectorEnv

from pyrlprob.utils.auxiliary import *
import pyrlprob.utils.callbacks as callbacks


def training(trainer: Union[str, Callable, Type], 
             config: Dict[str, Any], 
             stop: Dict[str, Any], 
             best_metric: str="episode_reward_mean",
             min_or_max: str="max",
             num_cp_to_keep: str="all",
             evaluation_active: bool=False,
             logdir: Optional[str]=None,
             create_out_file: bool=True, 
             load: Optional[Dict[str, Any]]=None, 
             debug: bool=False,
             open_ray: bool=True,
             return_time: bool=False) -> \
                 Union[Tuple[Dict[str, Any], str, str, int, float], \
                      Tuple[Dict[str, Any], str, str, int]]:
    """
    Train the current model with ray.tune, using the specified trainer and configs.

    Args:
        trainer (str or callable): trainer (i.e., RL algorithm) to train the model with
        config (dict): config file (dictionary)
        stop (dict): stopping conditions (dictionary)
        best_metric (str): metric to be used to determine the best result.
        min_or_max (str): if best_metric must be minimized or maximized
        num_cp_to_keep (str): number of checkpoints to keep. If "all", all checkpoints are kept, if "best", only the best checkpoint is kept
        evaluation_active (bool): whether evaluation is active
        logdir (str): name of the directory where training results are saved
        create_out_file (bool): whether to create an outfile with run time and best result
        load (dict): dictionary containing the directory and checkpoint where the 
            pre-trained model to load is located
        debug (bool): whether to print worker's logs.
        open_ray (bool): whether to open/close ray
        return_time (bool): whether to return run time per iter
    
    Return:
        best_result (str): best result of the experiment
        trainer_dir (str): trainer's output directory
        best_exp_dir (str): directory containing the best experiment's output
        last_checkpoint (int): last checkpoint saved
        (optional) run_time (float): total run time
    """

    # Create output folder
    if load is not None:
        outdir = load["logdir"]
        restore = load["checkpoint_dir"]
    else:
        if logdir is None:
            outdir = "./results/"
        else:
            outdir = logdir
        os.makedirs(outdir, exist_ok=True)
        restore = None

    # Initialize ray
    if open_ray:
        if debug:
            logging_level = "INFO"
        else:
            logging_level = "ERROR"
        ray.init(logging_level=logging_level, 
            log_to_driver=debug, 
            ignore_reinit_error=True)
    
    # Checkpoint saving
    if num_cp_to_keep == "all":
        keep_checkpoints_num = None
        checkpoint_score_attr = None
    elif num_cp_to_keep == "best":
        keep_checkpoints_num = 1
        if evaluation_active:
            checkpoint_score_attr = "evaluation/" + best_metric
        else:
            checkpoint_score_attr = best_metric
    else:
        keep_checkpoints_num = None
        checkpoint_score_attr = None

    # Train the model
    start_time = time.time()
    analysis = tune.run(trainer,
                        config=config,
                        local_dir=outdir,
                        restore=restore,
                        stop=stop,
                        metric=best_metric,
                        mode=min_or_max,
                        checkpoint_freq=1,
                        checkpoint_at_end=True,
                        keep_checkpoints_num=keep_checkpoints_num,
                        checkpoint_score_attr=checkpoint_score_attr)
    end_time = time.time()
    run_time = end_time - start_time

    # Last iteration stats
    best_result = analysis.best_result
    last_result = analysis.get_best_trial("training_iteration", "max").last_result
    best_exp_dir = analysis.best_logdir
    trainer_dir = best_exp_dir[:best_exp_dir.rfind("/")+1]
    best_exp_dir = best_exp_dir + "/"

    # Last checkpoint
    best_checkpoint_path = analysis.best_checkpoint
    if num_cp_to_keep == "best":
        last_checkpoint = int(best_checkpoint_path[best_checkpoint_path.rfind("-")+1:])
    else:
        last_checkpoint = last_result["training_iteration"]
    
    # Run time per iter
    run_time_per_iter = last_result["time_this_iter_s"]

    # Save elapsed time and results
    if create_out_file:
        f_out_res = open(best_exp_dir + "result.txt", "w")
        f_out_res.write("%22s %22s\n" \
            % ("# elapsed time [s]", "training_iteration"))
        f_out_res.write("%22.7f %22d\n" \
            % (run_time, best_result["training_iteration"]))
        f_out_res.close()

    # Terminate ray
    if open_ray:
        ray.shutdown()

    # Return trainer and best experiment directory + last checkpoint saved
    # and the run time per iter
    if return_time:
        return best_result, trainer_dir, best_exp_dir, last_checkpoint, run_time_per_iter
    else:
        return best_result, trainer_dir, best_exp_dir, last_checkpoint


def evaluation(trainer: Union[str, Callable, Type], 
               trainer_dir: str,
               exp_dirs: List[str],
               last_cps: List[int],
               config: Dict[str, Any],
               env_name: str, 
               env_config: Dict[str, Any],
               evaluation_num_episodes: int, 
               evaluation_config: Dict[str, Any],
               custom_eval_function: Optional[Union[Callable, str]]=None,
               best_metric: str="episode_reward_mean",
               min_or_max: str="max",
               metrics_and_data: Optional[Dict[str, Any]]=None,
               is_evaluation_env: bool=False,
               do_postprocess: bool=True,
               debug: bool=False) -> str:
    """
    Evaluate a model, and find the best checkpoint

    Args:
        trainer (str or callable): trainer (i.e., RL algorithm) with which the model was trained
        trainer_dir (str): trainer directory
        exp_dirs (list): list with experiments directories
        last_cps (list): list with last checkpoint number of each experiment in exp_dirs
        config (dict): config file (dictionary)
        env_name (str): environment class name
        env_config (dict): dictionary containing the environment configs
        evaluation_num_episodes (int): number of evaluation episodes
        evaluation_config (dict): dictionary containing the evaluation configs
        custom_eval_function (callable or str): Custom evaluation function (or function name)
        best_metric (str): metric to be used to determine the best checkpoint in exp_dirs.
            It it is a number, that checkpoint will be used.
        min_or_max (str): if best_metric must be minimized or maximized
        metrics_and_data (dict): dictionary containing the metrics and data to save
            in the new file progress.csv
        is_evaluation_env (bool): are metrics computed through an evaluation environment?
        do_postprocess (bool): whether to do postprocessing
        debug (bool):  whether to print data fro debugging
    """
    
    # Path of metrics
    metric_path = ""
    if is_evaluation_env:
        metric_path = "evaluation/"

    if trainer_dir is not None:
        # Check if best_metric is a number
        best_metric = str(best_metric)
        if best_metric.isdigit():
            best_cp = int(best_metric)
        else:
            # Get the best metric trend
            best_metric_trend = metric_training_trend(metric_path + best_metric,
                                                exp_dirs,
                                                last_cps)
            if min_or_max == "max":
                best_cp = np.argmax(best_metric_trend)+1
            else:
                best_cp = np.argmin(best_metric_trend)+1
        best_exp = next(exp for exp, cp in enumerate(last_cps) if cp >= best_cp)
        best_exp_dir = exp_dirs[best_exp]

        # Define load properties
        load = {}
        load["logdir"] = trainer_dir
        _, load["checkpoint_dir"] = \
            get_cp_dir_and_model(best_exp_dir, best_cp)
    else:
        load = None

    # Check what metrics/data are defined
    if metrics_and_data is None:
        metrics_and_data = {}
    for key in ["episode_step_data", "episode_end_data", "custom_metrics"]:
        if not key in metrics_and_data:
            metrics_and_data[key] = []      
                
    # Define trainer and configs for evaluation

    # Env config
    config["env"] = env_name
    mod_name, fun_name = env_name.rsplit('.',1)
    mod = importlib.import_module(mod_name)
    env = getattr(mod, fun_name)
    tune.register_env(config["env"], lambda config: env(config))
    config["env_config"] = env_config
    env_inst = env(env_config)
    if (issubclass(type(env_inst), VectorEnv)):
        env_inst = env_inst.env
        config["env"] = type(env_inst)

    # No learning
    config["lr"] = 0.

    # Evaluation and callbacks config
    config["create_env_on_driver"] = False
    config["evaluation_interval"] = 1
    config["evaluation_num_workers"] = max(config["evaluation_num_workers"], 1)
    if evaluation_num_episodes % config["num_envs_per_worker"]:
        if evaluation_num_episodes % config["evaluation_num_workers"]:
            config["evaluation_num_episodes"] = evaluation_num_episodes
            config["num_envs_per_worker"] = 1
            config["evaluation_num_workers"] = 1
        else:
            config["evaluation_num_episodes"] = evaluation_num_episodes
            config["num_envs_per_worker"] = 1
    else:
        if evaluation_num_episodes % config["evaluation_num_workers"]:
            config["evaluation_num_episodes"] = int(evaluation_num_episodes/config["num_envs_per_worker"])
            config["evaluation_num_workers"] = 1
        else:
            config["evaluation_num_episodes"] = int(evaluation_num_episodes/config["num_envs_per_worker"])
        
    config["evaluation_config"] = evaluation_config
    if custom_eval_function is not None:
        if callable(custom_eval_function):
                config["custom_eval_function"] = custom_eval_function
        else:
            mod_name, fun_name = custom_eval_function.rsplit('.',1)
            mod = importlib.import_module(mod_name)
            config["custom_eval_function"] = getattr(mod, fun_name)
    config["callbacks"] = callbacks.EvaluationCallbacks
    stop = {"training_iteration": 1}

    # Set the correct number of cpus/gpus for evaluation
    config["num_gpus_per_worker"] = 0
    config["num_gpus"] = 0
    total_w = config["num_workers"] + config["evaluation_num_workers"]
    cpus_count = config["num_cpus_per_worker"]*total_w + config["num_cpus_for_driver"]
    cpus_machine = multiprocessing.cpu_count()
    if cpus_count > cpus_machine:
        config["num_cpus_per_worker"] = (cpus_machine - 1.)/total_w if (cpus_machine - 1.)/total_w < 1 else int((cpus_machine - 1.)/total_w)
        config["num_cpus_for_driver"] = int(cpus_machine - config["num_cpus_per_worker"]*total_w)

    # Evaluation
    _, _, new_best_exp_dir, last_checkpoint  = training(trainer=trainer, 
                                                     config=config,
                                                     stop=stop,
                                                     load=load,
                                                     debug=debug)
    
    # Postprocessing
    if do_postprocess:
        cp_dir = postprocessing(best_exp_dir=new_best_exp_dir, 
                                checkpoint=last_checkpoint, 
                                metrics_and_data=metrics_and_data, 
                                is_evaluation_env=is_evaluation_env)
    else:
        cp_dir, _ = get_cp_dir_and_model(new_best_exp_dir, best_cp)
    
    return cp_dir


def postprocessing(best_exp_dir: str,
                   checkpoint: int,
                   metrics_and_data: Optional[Dict[str, Any]]=None,
                   is_evaluation_env: bool=False) -> str:
    """
    Default postprocessing.

    Args:
        exp_dir (str): experiment directory
        checkpoint (int): number of the checkpoint to postprocess
        custom_metrics (list): custom metrics to include in the postprocessing
        metrics_and_data (dict): dictionary containing the data and metrics to postprocess
        is_evaluation_env (bool): are metrics computed through an evaluation environment?
    """

    #Get checkpoint directory
    cp_dir, _ = get_cp_dir_and_model(best_exp_dir, checkpoint)

    #Path of metrics and data
    metric_path = ""
    if is_evaluation_env:
        metric_path = "evaluation/"

    if metrics_and_data is None:
        metrics_and_data = {}
        metrics_and_data["custom_metrics"] = []
        metrics_and_data["episode_step_data"] = []
        metrics_and_data["episode_end_data"] = []
    
    #Remove from metrics_and_data the metrics that are not in progress.csv
    for key in ["episode_step_data", "episode_end_data"]:
        metrics_and_data[key] = [metric for metric in metrics_and_data[key] if column_progress(best_exp_dir+"progress.csv", metric_path + "hist_stats/" + metric) != None]
    metrics_and_data["custom_metrics"] = [metric for metric in metrics_and_data["custom_metrics"] if \
                                          column_progress(best_exp_dir+"progress.csv", metric_path + "custom_metrics/" + metric + "_mean") != None]
    
    #Metrics and data to log
    values = ["min", "mean", "max"]
    metrics = {
        "metrics": {"episode_reward": {value: 0. for value in values}},
        "custom_metrics": {metric: {value: 0. for value in values} 
            for metric in metrics_and_data["custom_metrics"]},
        "episode_step_data": {metric: []
            for metric in metrics_and_data["episode_step_data"]}, 
        "episode_end_data": {metric: []
            for metric in metrics_and_data["custom_metrics"] + metrics_and_data["episode_end_data"]}
        }

    #Create output files
    f_log = open(cp_dir + "metrics.txt", "w")
    f_end_data = open(cp_dir + "episode_end_data.txt", "w")
    f_step_data = open(cp_dir + "episode_step_data.txt", "w")

    f_log.write("%20s " % ("# checkpoint"))
    for key, item in metrics.items():
        for key_in in item.keys():
            if key == "episode_step_data":
                f_step_data.write("%20s " % (key_in))
            elif key == "episode_end_data":
                f_end_data.write("%20s " % (key_in))
            else:
                for value in values:
                    f_log.write("%20s " % (key_in + "_" + value))
    f_log.write("\n")
    f_step_data.write("\n")
    f_end_data.write("\n")
    f_log.write("%20d " % (checkpoint))

    #Episode data lengths
    first_key = next(iter(metrics["episode_step_data"]))
    data_length = column_progress(best_exp_dir+"progress.csv", \
                        metric_path + "hist_stats/" + first_key + "_length")
    data_length = data_length[0].strip("[").strip("]").split(", ")
    data_length = np.array(data_length, dtype=int)

    #Retrieve metrics and data from progress.csv
    for key, item in metrics.items():
        for key_in in item.keys():
            if "episode" in key:
                q = column_progress(best_exp_dir+"progress.csv", metric_path + "hist_stats/" + key_in)                
                q = q[0].strip("[").strip("]").split(", ")
                if key == "episode_step_data":
                    start = 0
                    stop = 0
                    for e_num, e in enumerate(data_length):
                        stop = stop + e
                        metrics[key][key_in].append(q[start:stop])
                        start = start + e
                else:
                    metrics[key][key_in] = q
            else:
                if key == "custom_metrics":
                    metric_type = "custom_metrics/"
                else:
                    metric_type = ""
                for value in values:
                    metrics[key][key_in][value] = column_progress(best_exp_dir+"progress.csv", \
                        metric_path + metric_type + key_in + "_" + value)
                    f_log.write("%20.7f " % (metrics[key][key_in][value][-1]))
    f_log.write("\n")
    f_log.close()

    #Print episode_step_data
    for e_num, e in enumerate(data_length):
        for h in range(e):
            for key_in in metrics["episode_step_data"].keys():
                f_step_data.write("%20s " % (metrics["episode_step_data"][key_in][e_num][h]))
            f_step_data.write("\n")
        f_step_data.write("\n\n")
    f_step_data.close()

    #Print episode_end_data
    for e_num, _ in enumerate(data_length):
        for key_in in metrics["episode_end_data"].keys():
            f_end_data.write("%20s " % (metrics["episode_end_data"][key_in][e_num]))
        f_end_data.write("\n")
    f_end_data.close()

    print("\nResults printed\n")

    return cp_dir


import os
from sympy import divisors
import yaml
import ray
from pyrlprob.problem import RLProblem

def tune_workers_envs(config_file, cpus, gpus, envs, min_w, max_w):
    
    ray.init(logging_level="ERROR", log_to_driver=False)

    config = yaml.safe_load(open(config_file))

    # Tests
    if gpus > 0:
        hardware = ["cpu_only", "gpu_d_cpu_w"]
    else:
        hardware = ["cpu_only"]

    # Output file
    os.makedirs("./tuning", exist_ok=True)
    f_log = open("./tuning/cpu_times.txt", "w")
    f_log.write("%20s %20s %20s %20s %20s %20s %20s %20s %20s %20s\n" \
        % ("# hardware", "workers", "envs_per_worker", \
        "eval_workers", "eval_epis", "cpus_per_w", "gpus_per_w", "cpus_per_d", "gpus_per_d", "time[s]"))

    # Run simulation
    for h in hardware:
        for w in divisors(envs):
            
            if w > max_w or w < min_w:
                continue
            
            eval_w = w
            eval_epis = eval_w
            total_w = w + eval_w

            if h == "cpu_only":
                cpus_per_w = (cpus - 1.)/total_w if (cpus - 1.)/total_w < 1 else int((cpus - 1.)/total_w)
                cpus_per_d = int(cpus - cpus_per_w*total_w)
                gpus_per_w = 0
                gpus_per_d = 0
            elif h == "gpu_only":
                cpus_per_w = 0
                cpus_per_d = 0
                gpus_per_w = gpus/(total_w + 1)
                gpus_per_d = gpus_per_w
            elif h == "gpu_d_cpu_w":
                cpus_per_w = cpus/total_w if cpus/total_w < 1 else int(cpus/total_w)
                cpus_per_d = int(cpus - cpus_per_w*total_w)
                gpus_per_w = 0
                gpus_per_d = gpus
            elif h == "gpu_w_cpu_d":
                cpus_per_w = 0
                cpus_per_d = cpus
                gpus_per_w = gpus/total_w
                gpus_per_d = 0
            
            # Training config
            config["stop"]["training_iteration"] = 1
            config["config"]["num_workers"] = w
            config["config"]["num_envs_per_worker"] = int(envs / w)
            config["config"]["num_cpus_per_worker"] = cpus_per_w
            config["config"]["num_cpus_for_driver"] = cpus_per_d
            config["config"]["num_gpus_per_worker"] = gpus_per_w
            config["config"]["num_gpus"] = gpus_per_d
            config["config"]["create_env_on_driver"] = False

            #Evaluation config
            config["config"]["evaluation_interval"] = 1
            config["config"]["evaluation_num_episodes"] = eval_epis
            config["config"]["evaluation_num_workers"] = eval_w
            config["config"]["evaluation_config"]["explore"] = False

            # Define RL problem
            Prb = RLProblem(config)

            # Solve RL problem
            best_results, trainer_dir, exp_dirs, last_cps, best_cp_dir, run_time = \
                    Prb.solve(evaluate=False, postprocess=False, debug=False, 
                        open_ray=False, return_time=True)
            
            # Print results
            f_log.write("%20s %20d %20d %20d %20s %20.5f %20.5f %20.5f %20.5f %20.5f\n" \
                % (h, w, int(envs / w), eval_w, eval_epis, cpus_per_w, gpus_per_w, cpus_per_d, gpus_per_d, run_time))
            
            print("Done case: w = %d, cpu_per_w = %4.3f" % (w, cpus_per_w))
        
    f_log.close()

    ray.shutdown()

    return


# -*- coding: utf-8 -*-
from __future__ import print_function

import argparse
import sys
import sysconfig

from .commands import get_cmake_dir, get_include


def print_includes():
    # type: () -> None
    dirs = [
        sysconfig.get_path("include"),
        sysconfig.get_path("platinclude"),
        get_include(),
    ]

    # Make unique but preserve order
    unique_dirs = []
    for d in dirs:
        if d and d not in unique_dirs:
            unique_dirs.append(d)

    print(" ".join("-I" + d for d in unique_dirs))


def main():
    # type: () -> None

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--includes",
        action="store_true",
        help="Include flags for both pyrlprob and Python headers.",
    )
    parser.add_argument(
        "--cmakedir",
        action="store_true",
        help="Print the CMake module directory, ideal for setting -Dpyrlprob_ROOT in CMake.",
    )
    args = parser.parse_args()
    if not sys.argv[1:]:
        parser.print_help()
    if args.includes:
        print_includes()
    if args.cmakedir:
        print(get_cmake_dir())


if __name__ == "__main__":
    main()


from pyrlprob.problem import RLProblem
from pyrlprob.mdp import AbstractMDP
from pyrlprob.base_funs import training, evaluation, postprocessing

__all__ = [
    "RLProblem",
    "AbstractMDP",
    "training",
    "evaluation",
    "postprocessing",
]


import numpy as np
from typing import *

import gym
from gym import spaces

from ray.rllib.env.vector_env import VectorEnv

from pyrlprob.tests.cpp_tests.landing1d_cpp import Landing1DEnv_cpp, Landing1DVectorEnv_cpp


#Default config for Landing1DEnv
DEFAULT_CONFIG_LANDING1D = {
    "H": 40,
    "h0_min": 0.8,
    "h0_max": 1.2,
    "v0_min": -0.85,
    "v0_max": -0.75,
    "m0": 1.0,
    "tf": 1.397,
    "hf": 0.0,
    "vf": 0.0,
    "Tmax": 1.227,
    "c": 2.349,
    "g": 1.0
}


class cppLanding1DEnv(Landing1DEnv_cpp, gym.Env):
    """
    One-Dimensional Landing Problem.
    Reference: https://doi.org/10.2514/6.2008-6615
    """

    def __init__(self, 
                 config: Optional[Dict[str, Union[int, float]]] = None) -> None:
        """ 
        Class constructor 
        
        Args:
            config (dict): environment configs
        """

        if (not config) or (config is None):
            config = DEFAULT_CONFIG_LANDING1D

        Landing1DEnv_cpp.__init__(self, config)


class cppLanding1DVectorEnv(Landing1DVectorEnv_cpp, VectorEnv):
    """Vectorized version of the Landing1DEnv on multi-CPU processors.

    Contains `num_envs` Landing1DEnv instances.
    """

    def __init__(self, 
                 config: Optional[Dict[str, Union[int, float]]] = None) -> None:
        """ 
        Class constructor 
        
        Args:
            config (dict): environment configs
        """


        #Initialize parent classes
        Landing1DVectorEnv_cpp.__init__(
            self, 
            config["num_envs"], 
            config["num_threads"], 
            config)
        VectorEnv.__init__(
            self,
            observation_space=self.envs[0].observation_space,
            action_space=self.envs[0].action_space,
            num_envs=config["num_envs"])
        





from pyrlprob.utils.auxiliary import update, set_global_seeds, get_cp_dir_and_model, \
    moving_average, column_progress, metric_training_trend
from pyrlprob.utils.callbacks import TrainingCallbacks, epsConstraintCallbacks, EvaluationCallbacks
from pyrlprob.utils.plots import plot_metric

__all__ = [
    "update",
    "set_global_seeds",
    "get_cp_dir_and_model",
    "moving_average",
    "column_progress",
    "metric_training_trend",
    "TrainingCallbacks",
    "epsConstraintCallbacks",
    "EvaluationCallbacks",
    "plot_metric",
]


import gym

import numpy as np
from typing import *

from pyrlprob.utils.auxiliary import *

class AbstractMDP(gym.Env):
    """
    Abstract Markov Decision Process, based on OpenAI Gym.
    It encapsulates an environment (MDP) with
    arbitrary behind-the-scenes dynamics. The MDP can be
    partially or fully-observed. The control can differ from the returned action.
    An epsilon-constraint law is already implemented in this class,
    and can be called in derived classes.

    The main API methods that users of this class need to know and implement are:
        get_observation
        get_control
        next_state
        collect_reward
        get_info
        reset
        render (optional)

    And set the following attributes:
        action_space: the Space object corresponding to valid actions
        observation_space: the Space object corresponding to valid observations
        reward_range: a tuple corresponding to the min and max possible rewards (optional)
        time_step: length of any time-step in the MDP (optional)
        max_episode_steps: an int indicating the maximum number of steps in one episode (optional)
    
    If the user wants to use the callbacks defined in pyrlprob.utils.callbacks, the info dictionary
    returned by get_info should contain the following keys:
        custom_metrics: metrics to be included in tensorboard data
        episode_step_data: data returned and saved at each step
        episode_end_data: data returned and saved just at the end of the episode
    
    The epsilon-constraint technique can be activated just by including epsConstraintCallbacks 
    among the callbacks in the config file and by properly setting the
    class attributes self.epsilon0, self.epsilonf, self.iter0, self.iterf, which specify the initial and
    final value of the constraint satisfaction tolerance, and the training iterations which define
    the interval in which epsilon decreases.
    """

    def __init__(self, 
                 config = Dict[str, Any]) -> None:
        """ 
        Class constructor 
        
        Args:
            config (dict): environment configs
        """

        super().__init__()

        #Create class attributes
        for key, item in config.items():
            setattr(self, key, item)
        
        self.state = None
        self.time_step = None
        self.max_episode_steps = 999999
        self.done = False

        if hasattr(self, 'eps_schedule'):
            self.iter0 = self.eps_schedule[0][0]
            self.epsilon0 = self.eps_schedule[0][1]
            self.iterf = self.eps_schedule[1][0]
            self.epsilonf = self.eps_schedule[1][1]
        else:
            self.iter0 = 0
            self.epsilon0 = 0.
            self.iterf = 1
            self.epsilonf = 0.
        self.epsilon = self.epsilon0

        if hasattr(self, 'prng_seed'):
            seeds = self.seed(self.prng_seed)


    def seed(self,
             seed: Optional[int]=None) -> List[int]:
        """
        Set global seeds.

        Args:
            seed (int): seed
        """

        seeds = set_global_seeds(seed)

        return seeds
    

    def get_observation(self,
                        state: Optional[Any]=None,
                        control: Optional[Any]=None) -> Any:
        """
        Get current observation.

        Args:
            state (Any): current system state
            control (Any): current control action
        
        Return:
            observation (Any): current observation

        """

        raise NotImplementedError(
            "The Environment must implement the get_observation function!")
    

    def get_control(self,
                    action: Any,
                    state: Optional[Any]=None) -> Any:
        """
        Get current control.

        Args:
            action (Any): current action
            state (Any): current system state
        
        Return:
            control (Any): current control

        """

        raise NotImplementedError(
            "The Environment must implement the get_control function!")
    

    def next_state(self,
                   state: Optional[Any]=None, 
                   control: Optional[Any]=None,
                   time_step: Optional[Any]=None) -> Any:
        """
        Propagate state

        Args:
            state (Any): current system state
            control (Any): current control
            time_step (Any): time step for state transition
        
        Return:
            next_state (Any): next system state

        """

        raise NotImplementedError(
            "The Environment must implement the next_state function!")
    

    def collect_reward(self,
                       prev_state: Optional[Any]=None, 
                       state: Optional[Any]=None, 
                       control: Optional[Any]=None) -> Tuple[float, bool]:
        """
        Get current reward and done signal.

        Args:
            prev_state (Any): previous system state
            control (Any): current control
            state (Any): current system state
            
        Return:
            reward (float): current reward
            done (bool): is episode done?

        """

        raise NotImplementedError(
            "The Environment must implement the collect_reward function!")
    

    def get_info(self,
                 prev_state: Optional[Any]=None,
                 state: Optional[Any]=None,
                 observation: Optional[Any]=None,
                 control: Optional[Any]=None,
                 reward: Optional[float]=None,
                 done: Optional[bool]=None) -> Dict[str, Any]:
        """
        Get current info.

        Args:
            prev_state (Any): previous system state
            state (Any): current system state
            observation (Any): current observation
            control (Any): current control
            reward (float): last reward
            done (bool): last done signal
            
        Return:
            info (dict): current info

        """

        raise NotImplementedError(
            "The Environment must implement the get_info function!")
    

    def set_cstr_tolerance(self,
                           iter: int) -> None:
        """
        Set current constraint satisfation tolerance epsilon.

        Args:
            iter (int): current training iteration
        """

        if iter <= self.iter0:
            self.epsilon = self.epsilon0
        elif iter >= self.iterf:
            self.epsilon = self.epsilonf
        else:
            self.epsilon = \
                self.epsilon0*(self.epsilonf/self.epsilon0)**(float(iter - \
                self.iter0)/float(self.iterf - self.iter0))
    

    def step(self, action):
        """
        Step of the MDP.

        Args:
            action: current action
        """

        # Invalid action
        assert self.action_space.contains(action), "%r (%s) invalid"%(action, type(action))

        # Previous state
        self.prev_state = self.state

        # Get control
        control = self.get_control(action, self.prev_state)

        # Next state
        self.state = self.next_state(self.prev_state, control, self.time_step)

        # Get observation
        observation = self.get_observation(self.state, control)

        # Get reward and done signal
        reward, done = self.collect_reward(self.prev_state, self.state, control)
        self.done = done

        # Compute infos
        info = self.get_info(self.prev_state, self.state, observation, control, reward, done)

        return observation, float(reward), done, info
    

    def render(self, mode='human'):
        pass





    


try:
    from pyrlprob.tests.cpp_tests.landing1d import cppLanding1DEnv, cppLanding1DVectorEnv

    __all__ = [
        "cppLanding1DEnv",
        "cppLanding1DVectorEnv",
    ]
except ImportError:
    pass






import numpy as np
from numpy.random import uniform
from typing import *

import gym
from gym import spaces

from pyrlprob.tests.py_tests.landing1d_dyn import *


#Default config for Landing1DEnv
DEFAULT_CONFIG_LANDING1D = {
    "H": 40,
    "h0_min": 0.8,
    "h0_max": 1.2,
    "v0_min": -0.85,
    "v0_max": -0.75,
    "m0": 1.0,
    "tf": 1.397,
    "hf": 0.0,
    "vf": 0.0,
    "Tmax": 1.227,
    "c": 2.349,
    "g": 1.0
}


class pyLanding1DEnvGym(gym.Env):
    """
    One-Dimensional Landing Problem.
    Reference: https://doi.org/10.2514/6.2008-6615
    """

    def __init__(self, 
                 config: Optional[Dict[str, Any]] = None) -> None:
        """ 
        Class constructor 
        
        Args:
            config (dict): environment configs
        """

        super().__init__()

        if (not config) or (config is None):
            config = DEFAULT_CONFIG_LANDING1D

        #Create class attributes
        for key, item in config.items():
            setattr(self, key, item)

        #Class attributes
        self.dt = self.tf/(float(self.H))

        #Observation space
        self.observation_space = spaces.Box(low=-np.inf, high=np.inf, shape=(4,), dtype=np.float32)

        #Action space
        self.action_space = spaces.Box(low=-1., high=1., shape=(1,), dtype=np.float32)

        #Reward range
        self.reward_range = (-float('inf'), 0.)

        #Maximum episode steps
        self.max_episode_steps = self.H
    

    def get_observation(self,
                        state: Optional[Any]=None,
                        control: Optional[Any]=None) -> np.ndarray:
        """
        Get current observation: height and velocity
        """

        observation = np.array([state["h"], state["v"], state["m"], state["t"]], dtype=np.float32)

        return observation
    

    def get_control(self,
                    action: Any,
                    state: Optional[Any]=None) -> float:
        """
        Get current control: thrust value
        """

        control = 0.5*(action[0] + 1.)*self.Tmax

        return control


    def next_state(self,
                   state: Optional[Any], 
                   control: Any) -> Dict[str, float]:
        """
        Propagate state: integration of system dynamics
        """

        #State at current time-step
        s = np.array([state["h"], state["v"], state["m"]])

        #Integration of equations of motion
        data = np.array([self.g, control, self.c])

        # Integration
        t_eval = np.array([state['t'], state['t'] + self.dt])
        sol = rk4(dynamics, s, t_eval, data)

        #State at next time-step
        self.success = True
        s_new = {"h": sol[0], "v": sol[1], "m": sol[2], "t": t_eval[-1], "step": state["step"] + 1}
        
        return s_new
    

    def collect_reward(self,
                       prev_state: Optional[Any]=None, 
                       state: Optional[Any]=None, 
                       control: Optional[Any]=None) -> Tuple[float, bool]:
        """
        Get current reward and done signal.
        """

        done = False

        reward = state["m"] - prev_state["m"]

        if state["step"] == self.H:
            done = True
        if not self.success:
            done = True
        if state["h"] <= 0. or state["m"] <= 0.:
            done = True
        
        if done:
            cstr_viol = max(abs(state["h"] - self.hf), abs(state["v"] - self.vf) - 0.005)
            state["cstr_viol"] = max(cstr_viol, 0.)

            reward = reward - 10.*cstr_viol
        
        return reward, done
    

    def get_info(self,
                 prev_state: Optional[Any]=None,
                 state: Optional[Any]=None,
                 observation: Optional[Any]=None,
                 control: Optional[Any]=None,
                 reward: Optional[float]=None,
                 done: Optional[bool]=None) -> Dict[str, float]:
        """
        Get current info.
        """

        info = {}
        info["episode_step_data"] = {}
        info["episode_step_data"]["h"] = [prev_state["h"]] 
        info["episode_step_data"]["v"] = [prev_state["v"]] 
        info["episode_step_data"]["m"] = [prev_state["m"]] 
        info["episode_step_data"]["t"] = [prev_state["t"]] 
        info["episode_step_data"]["T"] = [control]
        if done:
            info["custom_metrics"] = {}
            info["episode_end_data"] = {}

            info["episode_step_data"]["h"].append(state["h"]) 
            info["episode_step_data"]["v"].append(state["v"]) 
            info["episode_step_data"]["m"].append(state["m"]) 
            info["episode_step_data"]["t"].append(state["t"]) 
            info["episode_step_data"]["T"].append(control)

            info["custom_metrics"]["cstr_viol"] = state["cstr_viol"]
            info["episode_end_data"]["hf"] = [state["h"]]
            info["episode_end_data"]["vf"] = [state["v"]]
            info["episode_end_data"]["mf"] = [state["m"]]

        return info
    

    def step(self, action):
        """
        Step of the MDP.

        Args:
            action: current action
        """

        # Invalid action
        assert self.action_space.contains(action), "%r (%s) invalid"%(action, type(action))

        # Previous state
        self.prev_state = self.state

        # Get control
        control = self.get_control(action, self.prev_state)

        # Next state
        self.state = self.next_state(self.prev_state, control)

        # Get observation
        observation = self.get_observation(self.state, control)

        # Get reward and done signal
        reward, done = self.collect_reward(self.prev_state, self.state, control)

        # Compute infos
        info = self.get_info(self.prev_state, self.state, observation, control, reward, done)

        return observation, float(reward), done, info
    

    def reset(self) -> np.ndarray:
        """ 
        Reset the environment
        """

        self.state = {}
        self.state["h"] = uniform(self.h0_min, self.h0_max)
        self.state["v"] = uniform(self.v0_min, self.v0_max)
        self.state["m"] = self.m0
        self.state["t"] = 0.
        self.state["step"] = 0

        control = 0.

        observation = self.get_observation(self.state, control)

        return observation
    

    def render(self, mode='human'):
        pass


    



import os
from typing import *
import numpy as np
import pandas as pd
import random
import gym
import tensorflow as tf
import torch
import collections.abc


def update(d, u):
    """
    Update the elements in dictionary d with those in dictionary u.
    """

    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


def set_global_seeds(seed: Optional[int]=None) -> List[int]:
    """
    Set the seed for python random, tensorflow, pytorch, numpy and gym spaces

    :param seed: (int) the seed
    """

    np.random.seed(seed)
    random.seed(seed)
    if hasattr(gym.spaces, 'prng'):
        gym.spaces.prng.seed(seed)

    return [seed]


def get_cp_dir_and_model(logdir: str, 
                         cp: int) -> Tuple[str, str]:
    """ 
    Get checkpoint directory and checkpoint model.

    Args:
        logdir (str): directory where the checkpoint is saved
        cp (int): checkpoint number
    
    Return:
        cp_dir (str): checkpoint directory
        cp_model (str): checkpoint model
    """

    cp_dir = logdir + "checkpoint_" + str(cp).zfill(6) + "/"
    cp_model = cp_dir + "checkpoint-" + str(cp)

    assert (os.path.exists(cp_dir)), "Folder %s does not exist!." % (cp_dir)

    return cp_dir, cp_model


def moving_average(values: np.ndarray,
                   window: int) -> np.ndarray:
    """
    Smooth values by doing a moving average

    Args:
        values (numpy array): actual values
        window (int): window's width
    
    Return:
        smoothed_values (numpy array): smoothed values
    """

    weights = np.repeat(1.0, window) / window
    smoothed_values = np.convolve(values, weights, 'valid')

    return smoothed_values


def column_progress(filename: str,
                    column: str, 
                    last_row: Optional[int] = None) -> List[Any]:
    """
    Return the list of values in a given column of a .csv file

    Args:
        filename (str): name of the .csv file with all the data
        column (str): header of the column of interest
        last_row (int): last raw of the column to process
    
    Return:
        column_data (list): list of values in the column
    """

    data = pd.read_csv(filename) 
    head = data.head(last_row)

    #Check if the column is the keys of head
    if column not in head.keys():
        return None
    
    column_data = list(head[column])

    return column_data


def metric_training_trend(metric_name: str,
                          experiment_dirs: List[str],
                          last_checkpoints: List[int]) -> List[Any]:
    """
    Retrieve the trend of a metric during training

    Args:
        metric_name (str): full name of the metric (with the path)
        experiment_dirs (list): list of the experiment directories
        last_checkpoints (list): list with the last checkpoints of the experiments
    
    Return:
        metric_trend (list): list with the values of the metric
    """

    check = column_progress(experiment_dirs[0]+"progress.csv", \
            metric_name, last_checkpoints[0])
    if check is None:
        return None
    
    metric_trend = []
    for exp_num, exp_dir in enumerate(experiment_dirs):
        metric_trend = metric_trend + column_progress(exp_dir+"progress.csv", \
            metric_name, last_checkpoints[exp_num])
    
    return metric_trend


from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(name='pyrlprob',
    version='2.2.8',
    author='Lorenzo Federici',
    author_email = 'federicilorenzo94@gmail.com',
    description = 'Train Gym-derived environments in Python/C++ through Ray RLlib',
    long_description = long_description,
    long_description_content_type = 'text/markdown',
    url = 'https://github.com/LorenzoFederici/pyrlprob',
    classifiers = [
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent'
    ],
    install_requires = [
        'dm-tree',
        'gym==0.15.3',
        'lz4',
        'matplotlib',
        'numpy==1.23.5',
        'pandas',
        'pybind11',
        'pytest',
        'ray==1.6.0',
        'scipy',
        'scikit-image',
        'opencv-python; platform_machine=="x86_64"',
        'tabulate',
        'tensorflow==2.15.0',
        'tensorboardx; platform_machine=="x86_64"',
        'torch',
        'typing',
        'PyYAML',
        'protobuf==3.20.3'
    ],
    packages = find_packages(),
    python_requires = '<3.10',
    include_package_data = True)


from pyrlprob.problem import RLProblem
from pyrlprob.tests import *

from typing import *
import matplotlib
import matplotlib.pyplot as plt
import os

from pyrlprob.utils.plots import plot_metric


def test_multiple_gym_envs(envs: List[str]=["CartPole-v1"], 
                           algs: List[List[str]]=[["PPO"]], 
                           res_dir: Optional[str]=None) -> None:
    """
    Compare multiple RL algorithms in different Gym environments through pyrlprob.

    Args:
        envs: list of Gym environments' names, among those in tuned_examples/
        algs: list of RL algorithms names, among those in tuned_examples/
        res_dir: path where results are saved. Current directory if not specified.
    """

    #Current location
    __location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))

    #Results dir
    if res_dir is None:
        res_dir = "./results/"

    #Plot style
    plt.style.use("seaborn")
    matplotlib.rc('font', size=24)
    palette = plt.rcParams['axes.prop_cycle'].by_key()['color']

    #For every env, do:
    for e, env in enumerate(envs):

        #Create res_dir
        res_dir_env = res_dir + env + "/"
        res_dir_plot = res_dir + "plots/"
        os.makedirs(res_dir_env, exist_ok=True)

        #Create figure
        fig = plt.figure()
        fig.set_size_inches(10.4,6.4)

        #For every alg, do:
        for a, alg in enumerate(algs[e]):
            
            #Config file
            config = os.path.join(__location__, "py_tests/tuned_examples/" + env + "_" + alg + ".yaml")

            #Problem definition
            Problem = RLProblem(config)

            #Training
            _, exp_dirs, last_cps, _ = \
                Problem.solve(res_dir_env, 
                              evaluate=False, 
                              postprocess=False)

            #Plot of metric trend
            fig = plot_metric("episode_reward",
                            exp_dirs,
                            last_cps,
                            fig=fig,
                            label=alg,
                            color=palette[a])
            ax = fig.gca()
            ax.tick_params(labelsize=24)
            ax.yaxis.grid(True, which='major', linewidth=0.5, linestyle='dashed')
            ax.xaxis.grid(True, which='major', linewidth=0.5, linestyle='dashed')
            plt.xlabel('Training iteration', fontsize=24)
            plt.ylabel('Episode return', fontsize=24)

        #Save figure
        ax.legend(fontsize=24, bbox_to_anchor=(0,1.01,1,0.2), loc="lower left", mode="expand", ncol=5)
        plt.tight_layout(pad=0)
        fig.savefig(res_dir_plot + env + "_reward.pdf", dpi=300)


if __name__ == "__main__":
    import pytest
    import sys
    sys.exit(pytest.main(["-v", __file__]))



